import numpy as np
from numba import jit, float64, int64
from scipy.sparse import csr_matrix

__all__ = ['calc_neighbourhood_density']


@jit(float64[:](int64[:], int64[:], float64[:], float64[:]), nopython=True)
def _indegree_calculator(ai: np.ndarray, aj: np.ndarray, v: np.ndarray, a: np.ndarray) -> np.ndarray:
    n_cells = a.shape[0]
    for i in range(n_cells):
        for p in range(ai[i], ai[i + 1]):
            j = aj[p]
            a[j] += v[p]
    return a


def calc_indegree(g: csr_matrix) -> np.ndarray:
    """
    Calculate indegree of each node in a graph

    Args:
        g: csr_matrix representing a graph

    Returns: Indegree of each node of a graph as a numpy vector
    """
    return _indegree_calculator(
        g.indptr.astype(np.int64), g.indices.astype(np.int64),
        g.data.astype(np.float64), np.zeros(g.shape[0]))


@jit(float64[:](int64[:], int64[:], float64[:]), nopython=True)
def _density_calculator(ai: np.ndarray, aj: np.ndarray, v: np.ndarray) -> np.ndarray:
    n_cells = v.shape[0]
    if n_cells != ai.shape[0] - 1:
        raise ValueError("ERROR: Vector size mismatch in `_iter_graph`. This is a bug. Please report.")
    a = np.zeros(n_cells)
    for i in range(n_cells):
        for p in range(ai[i], ai[i + 1]):
            j = aj[p]
            a[i] += v[j]
    return a + v


def calc_neighbourhood_density(g: csr_matrix, search_depth: int) -> np.ndarray:
    """
    Calculates the density of each node in the cell-cell neighbourhood graph.

    Args:
        g: csr_matrix representing a non-directed graph
        search_depth: If the value of `search_depth` is 0, then node density is simply the indegree of each node
                      in the graph. For values greater than 1 node density represents neighbourhood degree. For example,
                      when `search_depth` = 1, node density of a node is the sum of indegree of all its adjacent nodes
                      i.e. nodes that are at distance of 1. With increasing value of `neighbourhood_degree`, neighbours
                      further away from each node are also included included.

    Returns:

    """

    d = calc_indegree(g)
    ai, aj = g.indptr.astype(np.int64), g.indices.astype(np.int64)
    for n in range(search_depth):
        d = _density_calculator(ai, aj, d)
    return d



import networkx as nx
import numpy as np
from tqdm import tqdm


def make_digraph(d: np.ndarray, clust_info=None) -> nx.DiGraph:
    """
    Convert dendrogram into directed graph
    """

    g = nx.DiGraph()
    n = d.shape[0] + 1  # Dendrogram contains one less sample
    if clust_info is not None:
        if len(clust_info) != d.shape[0]+1:
            raise ValueError("ERROR: cluster information doesn't match number of leaves in dendrogram")
    else:
        clust_info = np.ones(d.shape[0]+1)*-1
    for i in tqdm(d, desc='Constructing graph from dendrogram'):
        v = i[2]  # Distance between clusters
        i = i.astype(int)
        g.add_node(n, nleaves=i[3], dist=v)
        if i[0] <= d.shape[0]:
            g.add_node(i[0], nleaves=0, dist=v, cluster=clust_info[i[0]])
        if i[1] <= d.shape[0]:
            g.add_node(i[1], nleaves=0, dist=v, cluster=clust_info[i[1]])
        g.add_edge(n, i[0])
        g.add_edge(n, i[1])
        n += 1
    if g.number_of_edges() != d.shape[0] * 2:
        raise ValueError('ERROR: Number of edges in directed graph not twice the dendrogram shape')
    return g


def find_lca(dag, start_leaf, target_n_leaves):
    tracker = start_leaf
    ancestor = None
    while True:
        p = list(dag.predecessors(tracker))
        if len(p) == 0:
            break
        p = p[0]
        if dag.nodes[p]['nleaves'] == target_n_leaves:
            ancestor = p
            break
        tracker = p
    return ancestor


def get_successors(dag, start):
    q = [start]
    d = []
    while len(q) > 0:
        i = q.pop(0)
        if dag.nodes[i]['nleaves'] == 0:
            d.append(i)
        else:
            q.extend(list(dag.successors(i)))
    return d


def recluster(dag, clust_members, target_n_clusts):
    lca = find_lca(dag, clust_members[0], clust_members.shape[0])
    if lca is None:
        raise ValueError("ERROR: LCA recognition failed")
    q = [lca]
    qvals = [1]
    clust_counts = 1
    clust_nodes = []
    while len(q) > 0:
        pop_index = np.argmax(qvals)
        p, _ = q.pop(pop_index), qvals.pop(pop_index)
        s = list(dag.successors(p))
        for i in s:
            if dag.nodes[i]['nleaves'] > 0:
                qvals.append(dag.nodes[i]['dist'])
                q.append(i)
            else:
                clust_nodes.append([i])
        clust_counts += 1
        if target_n_clusts <= clust_counts:
            break
    if sum([dag.nodes[x]['nleaves'] for x in q])+len(clust_nodes) == len(clust_members) is False:
        raise ValueError("ERROR: Incorrect number of leaves recovered")
    return [get_successors(dag, x) for x in q] + clust_nodes
